@@ -1,4 +1,4 @@
-
+// -*- test-case-name: nevow.test.test_javascript -*-
 // import Divmod
 // import Divmod.Base
 // import Divmod.Defer
@@ -24,109 +24,255 @@ Divmod.Runtime.NodeAttributeError.methods(
             " (programmer error).");
     });
 
+
+
+Divmod.Runtime.ScriptLoadingError = Divmod.Error.subclass('Divmod.Runtime.ScriptLoadingError');
+
+
+
+/**
+ * Wrapper around a lower-level scheduled timed call which provides additional
+ * useful operations.
+ */
+Divmod.Runtime.DelayedCall = Divmod.Class.subclass('Divmod.Runtime.DelayedCall');
+Divmod.Runtime.DelayedCall.methods(
+    function __init__(self, delay, callable) {
+        self._handle = setTimeout(callable, delay);
+    },
+
+    /**
+     * Unschedule this call so that it does not run.
+     */
+    function cancel(self) {
+        clearTimeout(self._handle);
+    });
+
+
+/**
+ * Move the given node so that it is a child of the given document and then try
+ * to find its node with the given id.  Restore the node to its original parent
+ * before returning the result.
+ */
+Divmod.Runtime._getElementByIdWithDocument = function _getElementByIdWithDocument(node, doc, id) {
+    var currentParent = node.parentNode;
+    var nextSibling = node.nextSibling;
+
+    // Insert ourselves into the document temporarily.
+    currentParent.removeChild(node);
+    doc.documentElement.appendChild(node);
+
+    // Try to find it.
+    var foundNode = node.ownerDocument.getElementById(id);
+
+    // And now put us back where we used to be.
+    node.parentNode.removeChild(node);
+    currentParent.insertBefore(node, nextSibling);
+
+    return foundNode;
+};
+
+/**
+ * XML parser for platforms which have the builtin "DOMParser" object.
+ */
+Divmod.Runtime._XMLParser = Divmod.Class.subclass("Divmod.Runtime._XMLParser");
+Divmod.Runtime._XMLParser.methods(
+    /**
+     * Return the cached DOM parser for this runtime.  Create one if one does
+     * not yet exist.
+     */
+    function _getParser(self) {
+        if (self._cachedParser === undefined) {
+            self._cachedParser = new DOMParser();
+        }
+        return self._cachedParser;
+    },
+
+    function parseXHTMLString(self, s) {
+        var doc = self._getParser().parseFromString(s, "application/xml");
+        var uri = doc.documentElement.namespaceURI;
+        if (uri != "http://www.w3.org/1999/xhtml") {
+            throw new Error(
+                "Unknown namespace (" + uri +
+                ") used with parseXHTMLString" +
+                "- only XHTML 1.0 is supported.");
+        }
+        return doc;
+    });
+
+
+
 Divmod.Runtime.Platform = Divmod.Class.subclass("Divmod.Runtime.Platform");
 
 Divmod.Runtime.Platform.DOM_DESCEND = 'Divmod.Runtime.Platform.DOM_DESCEND';
 Divmod.Runtime.Platform.DOM_CONTINUE = 'Divmod.Runtime.Platform.DOM_CONTINUE';
 Divmod.Runtime.Platform.DOM_TERMINATE = 'Divmod.Runtime.Platform.DOM_TERMINATE';
 
+/**
+ * The following constants are _supposed_ to be provided by the XHR object
+ * itself, as per http://www.w3.org/TR/XMLHttpRequest/#done
+ *
+ * However, no current browser (Firefox 2, IE6, IE7, Opera, Safari) actually
+ * provides these values, so we provide them here to avoid hard-coding
+ * integers.
+ */
+
+Divmod.Runtime.Platform.XHR_UNSENT = 0;
+Divmod.Runtime.Platform.XHR_OPEN = 1;
+Divmod.Runtime.Platform.XHR_SENT = 2;
+Divmod.Runtime.Platform.XHR_LOADING = 3;
+Divmod.Runtime.Platform.XHR_DONE = 4;
+
+
+/**
+ * A mapping of instance C{__id__}s to instances, used by
+ * Divmod.Runtime.Platform.connectSingleDOMEvent
+ */
+Divmod.Runtime._eventHandlerObjects = {};
+
+
+/**
+ * Make a DOM event handler for the L{Divmod.Class} instance with
+ * C{handlerObjectID}, which calls C{methodName} on the corresponding instance
+ * from L{Divmod.Runtime._eventHandlerObjects}, and then removes itself.
+ *
+ * This is a top level function (rather than a part of
+ * L{Divmod.Runtime.Platform.connectSingleDOMEvent}) to avoid closing over any
+ * DOM nodes, or anything which might hold a reference to one.  IE leaks
+ * memory when you do that.
+ *
+ * @see http://www.bazon.net/mishoo/articles.epl?art_id=824
+ */
+Divmod.Runtime._makeSingleEventHandler = function _makeSingleEventHandler(
+    handlerObjectID, domEventName, methodName) {
+
+    return function(node) {
+        delete node[domEventName];
+
+        var handlerObject = Divmod.Runtime._eventHandlerObjects[
+            handlerObjectID];
+
+        delete Divmod.Runtime._eventHandlerObjects[handlerObjectID];
+
+        return handlerObject[methodName].call(handlerObject, node);
+    };
+};
+
+/**
+ * @type _loadEventDelay: Number
+ * @ivar: The number of milliseconds by which to delay invocation of events
+ *     registered with C{addLoadEvent}.
+ */
 Divmod.Runtime.Platform.methods(
     function __init__(self, name) {
         self.name = name;
         self.attrNameToMangled = {};
-        self._scriptCounter = 0;
-        self._scriptDeferreds = {};
+        self._loadEventDelay = 1;
     },
 
-    function _nsResolver(self, prefix) {
-        var ns;
-        switch(prefix) {
-            case 'html':
-                ns = 'http://www.w3.org/1999/xhtml';
-                break;
-            case 'athena':
-                ns = 'http://divmod.org/ns/athena/0.7';
-                break;
-            default:
-                // this should never happen, but browsers still suck...
-                ns = null;
-        }
-        return ns;
-    },
+    /**
+     * Attach a DOM event handler to C{domNode} for the event C{domEventName}
+     * which will call the method named C{methodName} on the object
+     * C{handlerObject}.  The event handler will be removed from the node
+     * after the event is triggered.
+     *
+     * @param domEventName: the name of an event in the DOM, such as
+     * "onclick", "onscroll", "onchange", etc.
+     * @type domEventName: C{String}
+     *
+     * @param handlerObject: The object on which call C{methodName}.
+     * @type handlerObject: L{Divmod.Class}
+     *
+     * @param domNode: The node to attach the event handler to.
+     *
+     * @param methodName: The method name to call on C{handlerObject} when the
+     * event fires.
+     *
+     * @rtype: C{undefined}
+     */
+    function connectSingleDOMEvent(
+        self, domEventName, handlerObject, domNode, methodName) {
 
-    function _xpathNodeByAttribute(self, attrName, attrValue) {
-        var upper = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
-        var lower = upper.toLowerCase();
-        return (
-            ".//*[@*[translate(name(),'" + upper + "', '" + lower + "')='" +
-            attrName + "']='" + attrValue + "'] | .[@*[translate(name(),'" +
-            upper + "', '" + lower + "')='" + attrName + "']='" + attrValue +
-            "']");
+        Divmod.Runtime._eventHandlerObjects[
+            handlerObject.__id__] = handlerObject;
+
+        domNode[domEventName] = Divmod.Runtime._makeSingleEventHandler(
+            handlerObject.__id__, domEventName, methodName);
     },
 
+    /**
+     * Locate the first node with a given value for a particular attribute.
+     *
+     * @type node: Node
+     * @param node: The node from which to begin searching downward.
+     *
+     * @type attrName: String
+     * @param attrName: The name of the attribute to check.
+     *
+     * @type attrValue: String
+     * @param attrValue: The value of the named attribute for which to search.
+     *
+     * @rtype: Node
+     * @return: The first node found with a value matching the given one.
+     *
+     * @throw Error: If there is no node with an attribute which matches the
+     * given value.
+     */
     function firstNodeByAttribute(self, root, attrName, attrValue) {
         /* duplicate this here rather than adding an "onlyOne" arg to
            nodesByAttribute so adding an extra arg accidentally doesn't change
            it's behaviour if called directly
         */
-        var xpath = self._xpathNodeByAttribute(attrName, attrValue);
-        var node = document.evaluate(
-            xpath,
+        var descend = Divmod.Runtime.Platform.DOM_DESCEND;
+        var terminate = Divmod.Runtime.Platform.DOM_TERMINATE;
+
+        var result = null;
+        self.traverse(
             root,
-            null,
-            XPathResult.FIRST_ORDERED_NODE_TYPE,
-            null
-        ).singleNodeValue;
-        if (!node) {
+            function(node) {
+                if (self.getAttribute(node, attrName) == attrValue) {
+                    result = node;
+                    return terminate;
+                }
+                return descend;
+            });
+        if (result === null) {
             throw Divmod.Runtime.NodeAttributeError(root, attrName, attrValue);
         }
-        return node;
-    },
-
-    function nodeByAttribute(self, root, attrName, attrValue, /* optional */ defaultNode){
-        var xpath = self._xpathNodeByAttribute(attrName, attrValue);
-        var nodes = document.evaluate(
-            xpath,
-            root,
-            null,
-            XPathResult.UNORDERED_NODE_SNAPSHOT_TYPE,
-            null
-        );
-        if (nodes.snapshotLength > 1) {
-            throw new Error("Found too many " + attrName + " = " + attrValue);
-        }
-        else if (nodes.snapshotLength < 1) {
-            if (defaultNode === undefined) {
-                throw Divmod.Runtime.NodeAttributeError(root, attrName, attrValue);
-            }
-            else {
-                return defaultNode;
-            }
-        }
-        else {
-            var result = nodes.snapshotItem(0);
-            return result;
-        }
+        return result;
     },
 
     function nodesByAttribute(self, root, attrName, attrValue) {
+        var descend = Divmod.Runtime.Platform.DOM_DESCEND;
         var results = [];
-        var xpath = self._xpathNodeByAttribute(attrName, attrValue);
-        var nodes = document.evaluate(
-            xpath,
+        self.traverse(
             root,
-            null,
-            XPathResult.ORDERED_NODE_ITERATOR_TYPE,
-            null
-        );
-        var node = nodes.iterateNext();
-        while(node){
-            results.push(node);
-            node = nodes.iterateNext();
-        }
+            function(node) {
+                if (self.getAttribute(node, attrName) == attrValue) {
+                    results.push(node);
+                }
+                return descend;
+            });
         return results;
     },
 
+    function nodeByAttribute(self, root, attrName, attrValue, /* optional */ defaultNode) {
+        var nodes = self.nodesByAttribute(root, attrName, attrValue);
+        if (nodes.length > 1) {
+            throw new Error("Found too many " + attrName + " = " + attrValue);
+        } else if (nodes.length < 1) {
+            if (defaultNode === undefined) {
+                throw Divmod.Runtime.NodeAttributeError(
+                    root, attrName, attrValue);
+            } else {
+                return defaultNode;
+            }
+
+        } else {
+            var result = nodes[0];
+            return result;
+        }
+    },
+
     /**
      * Calculate the X/Y coordinates of event C{event} on the
      * page in a cross-browser fashion.  If C{event} is not defined,
@@ -273,15 +419,27 @@ Divmod.Runtime.Platform.methods(
         return localName;
     },
 
-//     function setAttribute(self, node, localName, namespaceURI, namespaceIdentifier, value) {
-//         if (namespaceURI !== undefined && node.setAttributeNS) {
-//             node.setAttributeNS(namespaceURI, localName, value);
-//         } else if (namespaceIdentifier !== undefined) {
-//             node.setAttribute(namespaceIdentifier + ':' + localName, value);
-//         } else {
-//             node.setAttribute(self._mangleAttributeName(localName), value);
-//         }
-//     },
+    /**
+     * Reliably set the value for a node attribute.
+     */
+    function setAttribute(self, node, localName, value, namespaceURI, namespaceIdentifier) {
+        if (namespaceURI === undefined && namespaceIdentifier === undefined) {
+            localName = self._mangleAttributeName(localName);
+        }
+        if (node.hasAttributeNS) {
+            if (node.hasAttributeNS(namespaceURI, localName)) {
+                return node.setAttributeNS(namespaceURI, localName, value);
+            } else if (node.hasAttributeNS(namespaceIdentifier, localName)) {
+                return node.setAttributeNS(namespaceIdentifier, localName, value);
+            }
+        }
+        if (node.setAttribute) {
+            var a = (namespaceIdentifier === undefined) ?
+                localName : namespaceIdentifier + ':' + localName;
+            return node.setAttribute(a, value);
+        }
+        return null;
+    },
 
     /**
      * This is _the_way_ to get the value of an attribute off of node
@@ -329,30 +487,74 @@ Divmod.Runtime.Platform.methods(
         throw new Error("makeHTTPRequest is unimplemented on " + self);
     },
 
-    function _onReadyStateChange(self, req, d) {
-        return function() {
-            if (d == null) {
-                /* We've been here before, and finished everything we needed
-                 * to finish.
-                 */
-                return;
-            }
-            if (typeof Divmod == 'undefined') {
-                /*
-                 * If I am invoked _after_ onunload is fired, the JS
-                 * environment has been torn down, and there is basically
-                 * nothing useful that any callbacks could do.  You can detect
-                 * this environment brokenness by looking at a top-level module
-                 * object (such as our own, Divmod).
-                 *
-                 * I also eliminate myself as the onreadystatechange handler of
-                 * this request, since at no future point will the execution
-                 * context magically be restored to a working state.
-                 */
-                req.onreadystatechange = null;
-                return;
-            }
-            if (req.readyState == 4) {
+    /**
+     * Asynchronously retrieve an HTTP resource.
+     *
+     * @param url: a string; the relative path of an URL.
+     *
+     * @param args: optional; an array of arrays (key-value pairs)
+     * representing query arguments to add to the URL.  defaults to [].
+     *
+     * @param action: optional; the HTTP method to use.  defaults to 'GET'.
+     *
+     * @param headers: optional; an array of arrays (key-value pairs)
+     * representing HTTP headers to set in the request.  defaults to [].
+     *
+     * @param content: optional; the payload of the HTTP request.  defaults to
+     * ''.
+     *
+     * @return: an array with 2 elements.  The first is an XMLHttpRequest
+     * object, whose API is browser-dependent but bears at least a passing
+     * resemblance to http://www.w3.org/TR/XMLHttpRequest/.  The second is a
+     * Deferred, which will fire when the request has completed, with an
+     * object that has 2 attributes: 'status' and 'response', or errback if
+     * the request fails due to a network error before a status has been
+     * retrieved.  The 'status' will be an integer, giving the HTTP status
+     * code of the response.  For example, 200 if successful.  The 'response'
+     * will be a string, the text of the response.
+     */
+    function getPage(self, url, /* optional */ args, action, headers, content, synchronous) {
+        // Fill out defaults.
+        if (args === undefined) {
+            args = [];
+        }
+        if (action === undefined) {
+            action = 'GET';
+        }
+        if (headers === undefined) {
+            headers = [];
+        }
+        if (content === undefined) {
+            content = '';
+        }
+        if (synchronous === undefined) {
+            synchronous = false;
+        }
+
+        // Construct URL by quoting and appending query arguments.
+        var qargs = [];
+        for (var i = 0; i < args.length; ++i) {
+            /* TODO: encodeURIComponent is not present on Safari, according to
+             * http://aptana.com/reference/api/Global.html: we'll have to
+             * update this to support it.
+             */
+            qargs.push(args[i][0] + '=' + encodeURIComponent(args[i][1]));
+        }
+        if (qargs.length) {
+            url = url + '?' + qargs.join('&');
+        }
+
+        // Build request with appropriate headers.
+        var req = self.makeHTTPRequest();
+        req.open(action, url, !synchronous);
+        for (var i = 0; i < headers.length; ++i) {
+            req.setRequestHeader(headers[i][0], headers[i][1]);
+        }
+
+        // Set up a callback to fire a deferred.
+        var d = new Divmod.Defer.Deferred();
+        req.onreadystatechange = function() {
+            if (req.readyState == Divmod.Runtime.Platform.XHR_DONE) {
                 var result = null;
                 try {
                     result = {'status': req.status,
@@ -363,47 +565,9 @@ Divmod.Runtime.Platform.methods(
                 if (result != null) {
                     d.callback(result);
                 }
-                d = null;
             }
         };
-    },
 
-    function getPage(self, url, /* optional */ args, action, headers, content) {
-        if (args == undefined) {
-            args = [];
-        }
-        if (action == undefined) {
-            action = 'GET';
-        }
-        if (headers == undefined) {
-            headers = [];
-        }
-        if (content == undefined) {
-            content = null;
-        }
-
-        var qargs = [];
-        for (var i = 0; i < args.length; ++i) {
-            /* encodeURIComponent may not exist on some browsers, I guess.  FF
-             * 1.5 and IE 6 have it, anyway.
-             */
-            qargs.push(args[i][0] + '=' + encodeURIComponent(args[i][1]));
-        }
-
-        if (qargs.length) {
-            url = url + '?' + qargs.join('&');
-        }
-
-        var d = new Divmod.Defer.Deferred();
-        var req = self.makeHTTPRequest();
-
-        req.open(action, url, true);
-
-        for (var i = 0; i < headers.length; ++i) {
-            req.setRequestHeader(headers[i][0], headers[i][1]);
-        }
-
-        req.onreadystatechange = self._onReadyStateChange(req, d);
         req.send(content);
         return [req, d];
     },
@@ -468,48 +632,16 @@ Divmod.Runtime.Platform.methods(
         self.appendNodeContent(node, innerHTML);
     },
 
-    /**
-     * Load the JavaScript module at the given URL.
-     *
-     * The manner in which the JavaScript is evaluated is implementation
-     * dependent.
-     *
-     * @return: A Deferred which fires when the contents of the module have
-     * been evaluated.
-     */
     function loadScript(self, location) {
-        self._scriptCounter += 1;
-        self._scriptDeferreds[self._scriptCounter] = Divmod.Defer.Deferred();
-
-        var language = '"text/javascript"';
-        var onload = '"Divmod.Runtime.theRuntime._scriptLoaded(' + self._scriptCounter + ')"';
-        var src = '"' + location + '"';
-        var id = '"__athena_runtime_script_loader_' + self._scriptCounter + '__"';
-        var xmlns = '"http://www.w3.org/1999/xhtml"';
-        var script = (
-            '<span ' +
-            'style="display: none" ' +
-            'xmlns=' + xmlns + ' ' +
-            'id=' + id + '>' +
-            '<script ' +
-            'type=' + language + ' ' +
-            'onload=' + onload + ' ' +
-            'src=' + src + '>' +
-            '</script>' +
-            '</span>');
-
-        self.appendNodeContent(document.body, script);
-
-        return self._scriptDeferreds[self._scriptCounter];
-    },
-
-    function _scriptLoaded(self, which) {
-        var script = document.getElementById('__athena_runtime_script_loader_' + which + '__');
-        script.parentNode.removeChild(script);
-
-        var loaded = self._scriptDeferreds[which];
-        delete self._scriptDeferreds[which];
-        loaded.callback(null);
+        // <script> tricks produce spectacularly bizarre behaviour in IE and
+        // Safari doesn't support onerror, so we just use getPage/eval here.
+        var req = Divmod.Runtime.theRuntime.getPage(location);
+        var d = req[1];
+        d.addCallback(
+            function (result) {
+                eval(result['response']);
+            });
+        return d;
     },
 
     /**
@@ -538,9 +670,128 @@ Divmod.Runtime.Platform.methods(
             throw Divmod.Runtime.NodeNotFound('Node with id ' + id + ' not found');
         }
         return foundNode;
+    },
+
+    /**
+     * Arrange for the given function to be called when the page is fully
+     * loaded.  Due to the behavior of various browsers (particularly with
+     * respect to their "page loading" aka "throbber" user interface element)
+     * this callable is not invoked directly in response to the DOM page load
+     * event, but in a delayed call scheduled from that event.
+     */
+    function addLoadEvent(self, callable) {
+        Divmod.Base.addLoadEvent(function() {
+                setTimeout(callable, self._loadEventDelay);
+            });
     });
 
-Divmod.Runtime.Firefox = Divmod.Runtime.Platform.subclass('Divmod.Runtime.Firefox');
+
+Divmod.Runtime.Spidermonkey = Divmod.Runtime.Platform.subclass(
+    'Divmod.Runtime.Spidermonkey');
+Divmod.Runtime.Spidermonkey.isThisTheOne = function isSpidermonkeyTheOne() {
+    try {
+        window;
+    } catch (error) {
+        return true;
+    }
+    return false;
+};
+
+
+Divmod.Runtime.XPathSupportingPlatform = Divmod.Runtime.Platform.subclass(
+    'Divmod.Runtime.XPathSupportingPlatform');
+Divmod.Runtime.XPathSupportingPlatform.methods(
+    function _nsResolver(self, prefix) {
+        var ns;
+        switch(prefix) {
+            case 'html':
+                ns = 'http://www.w3.org/1999/xhtml';
+                break;
+            case 'athena':
+                ns = 'http://divmod.org/ns/athena/0.7';
+                break;
+            default:
+                // this should never happen, but browsers still suck...
+                ns = null;
+        }
+        return ns;
+    },
+
+    function _xpathNodeByAttribute(self, attrName, attrValue) {
+        var upper = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
+        var lower = upper.toLowerCase();
+        return (
+            ".//*[@*[translate(name(),'" + upper + "', '" + lower + "')='" +
+            attrName + "']='" + attrValue + "'] | .[@*[translate(name(),'" +
+            upper + "', '" + lower + "')='" + attrName + "']='" + attrValue +
+            "']");
+    },
+
+    function firstNodeByAttribute(self, root, attrName, attrValue) {
+        /* duplicate this here rather than adding an "onlyOne" arg to
+           nodesByAttribute so adding an extra arg accidentally doesn't change
+           it's behaviour if called directly
+        */
+        var xpath = self._xpathNodeByAttribute(attrName, attrValue);
+        var node = document.evaluate(
+            xpath,
+            root,
+            null,
+            XPathResult.FIRST_ORDERED_NODE_TYPE,
+            null
+        ).singleNodeValue;
+        if (!node) {
+            throw Divmod.Runtime.NodeAttributeError(root, attrName, attrValue);
+        }
+        return node;
+    },
+
+    function nodeByAttribute(self, root, attrName, attrValue, /* optional */ defaultNode){
+        var xpath = self._xpathNodeByAttribute(attrName, attrValue);
+        var nodes = document.evaluate(
+            xpath,
+            root,
+            null,
+            XPathResult.UNORDERED_NODE_SNAPSHOT_TYPE,
+            null
+        );
+        if (nodes.snapshotLength > 1) {
+            throw new Error("Found too many " + attrName + " = " + attrValue);
+        }
+        else if (nodes.snapshotLength < 1) {
+            if (defaultNode === undefined) {
+                throw Divmod.Runtime.NodeAttributeError(root, attrName, attrValue);
+            }
+            else {
+                return defaultNode;
+            }
+        }
+        else {
+            var result = nodes.snapshotItem(0);
+            return result;
+        }
+    },
+
+    function nodesByAttribute(self, root, attrName, attrValue) {
+        var results = [];
+        var xpath = self._xpathNodeByAttribute(attrName, attrValue);
+        var nodes = document.evaluate(
+            xpath,
+            root,
+            null,
+            XPathResult.ORDERED_NODE_ITERATOR_TYPE,
+            null
+        );
+        var node = nodes.iterateNext();
+        while(node){
+            results.push(node);
+            node = nodes.iterateNext();
+        }
+        return results;
+    });
+
+Divmod.Runtime.Firefox = Divmod.Runtime.XPathSupportingPlatform.subclass(
+    'Divmod.Runtime.Firefox');
 
 Divmod.Runtime.Firefox.isThisTheOne = function isThisTheOne() {
     try {
@@ -557,8 +808,9 @@ Divmod.Runtime.Firefox.isThisTheOne = function isThisTheOne() {
 Divmod.Runtime.Firefox.methods(
     function __init__(self) {
         Divmod.Runtime.Firefox.upcall(self, '__init__', 'Firefox');
-        self.dp = new DOMParser();
-        self.ds = new XMLSerializer();
+        self._xmlparser = Divmod.Runtime._XMLParser();
+        self._scriptCounter = 0;
+        self._scriptDeferreds = {};
     },
 
     function makeHTML(self, element) {
@@ -586,11 +838,7 @@ Divmod.Runtime.Firefox.methods(
     },
 
     function parseXHTMLString(self, s) {
-        var doc = self.dp.parseFromString(s, "application/xml");
-        if (doc.documentElement.namespaceURI != "http://www.w3.org/1999/xhtml") {
-            throw new Error("Unknown namespace used with parseXHTMLString - only XHTML 1.0 is supported.");
-        }
-        return doc;
+        return self._xmlparser.parseXHTMLString(s);
     },
 
     function appendNodeContent(self, node, innerHTML) {
@@ -623,6 +871,147 @@ Divmod.Runtime.Firefox.methods(
         node.appendChild(document.importNode(doc.documentElement, true));
     },
 
+    function makeHTTPRequest(self) {
+        return new XMLHttpRequest();
+    },
+
+    /**
+     * Load the JavaScript module at the given URL.
+     *
+     * The manner in which the JavaScript is evaluated is implementation
+     * dependent.
+     *
+     * @return: A Deferred which fires when the contents of the module have
+     * been evaluated.
+     */
+    function loadScript(self, location) {
+        self._scriptCounter += 1;
+        var scriptID = '__athena_runtime_script_loader_' + self._scriptCounter + '__';
+        self._scriptDeferreds[scriptID] = Divmod.Defer.Deferred();
+
+        var language = '"text/javascript"';
+        var singleQuotedID = "'" + scriptID + "'";
+        var doubleQuotedID = '"' + scriptID + '"';
+        var onload = '"Divmod.Runtime.theRuntime._scriptLoaded(' + singleQuotedID + ')"';
+        var onerror = '"Divmod.Runtime.theRuntime._scriptError(' + singleQuotedID + ', arguments[0])"';
+        var src = '"' + location + '"';
+        var xmlns = '"http://www.w3.org/1999/xhtml"';
+        var script = (
+            '<span ' +
+            'style="display: none" ' +
+            'xmlns=' + xmlns + ' ' +
+            'id=' + doubleQuotedID + '>' +
+            '<script ' +
+            'type=' + language + ' ' +
+            'onload=' + onload + ' ' +
+            'onerror=' + onerror + ' ' +
+            'src=' + src + '>' +
+            '</script>' +
+            '</span>');
+
+        self.appendNodeContent(document.body, script);
+
+        return self._scriptDeferreds[scriptID];
+    },
+
+    function _scriptLoaded(self, scriptID) {
+        var script = document.getElementById(scriptID);
+        script.parentNode.removeChild(script);
+
+        var loaded = self._scriptDeferreds[scriptID];
+        delete self._scriptDeferreds[scriptID];
+        loaded.callback(null);
+    },
+
+    function _scriptError(self, scriptID, error) {
+        var script = document.getElementById(scriptID);
+        script.parentNode.removeChild(script);
+
+        var loaded = self._scriptDeferreds[scriptID];
+        delete self._scriptDeferreds[scriptID];
+        loaded.errback(Divmod.Runtime.ScriptLoadingError(error));
+    });
+
+
+Divmod.Runtime.WebKit = Divmod.Runtime.Platform.subclass(
+    'Divmod.Runtime.WebKit');
+
+Divmod.Runtime.WebKit.isThisTheOne = function isThisTheOne() {
+    try {
+        return navigator.userAgent.indexOf('WebKit') != -1;
+    } catch (err) {
+        if (err instanceof ReferenceError) {
+            return false;
+        } else {
+            throw err;
+        }
+    }
+};
+
+Divmod.Runtime.WebKit.methods(
+    function __init__(self) {
+        Divmod.Runtime.WebKit.upcall(self, '__init__', 'WebKit');
+        self._xmlparser = Divmod.Runtime._XMLParser();
+        // WebKit has no equivalent to the stacktrace that FF provides, so this
+        // JSON adapter will provide a dummy object to make Athena happy when
+        // it tries to send exceptions from the client to the server
+        Divmod.Base.registerJSON(
+            'Error',
+            function(obj) {
+                return arguments[1] instanceof Error;
+            },
+            function(obj) {
+                var exc = arguments[1];
+                return {
+                    'name': exc.name,
+                    'message': exc.message,
+                    'stack': 'No stacktrace available\n'
+                };
+            }
+        );
+        /*
+         * WebKit fires the onload event long before the page has actually
+         * loaded.  This has two consequences.  First, if the Athena transport
+         * is started when the onload event fires, then WebKit will believe
+         * that the page never actually completes loading, since there will be
+         * an open, long-lived XHR.  This means its spinner will never stop
+         * spinning, which will aggravate users.  Second, if the Athena
+         * transport is started before the page fully loads, application-level
+         * JavaScript may fail, since the page is supposed to have loaded by
+         * the time that event fires.  This delay is a complete and utter hack
+         * to try to work around these two problems in *most* cases.  It will
+         * certainly fail in some cases.  Unless WebKit exposes a real event
+         * which fires when the page is really loaded, though, I don't see any
+         * easy way around this. -exarkun
+         */
+        self._loadEventDelay = 200;
+    },
+
+    function getElementByIdWithNode(self, node, id) {
+        var foundNode = node.ownerDocument.getElementById(id);
+        if (foundNode === null) {
+            // We didn't find it, maybe we need a workaround.
+            // Let's insert ourselves into the document temporarily.
+            foundNode = Divmod.Runtime._getElementByIdWithDocument(
+                node, document, id);
+        }
+
+        if (foundNode === null) {
+            throw Divmod.Runtime.NodeNotFound('Node with id ' + id + ' not found');
+        }
+
+        return foundNode;
+    },
+
+    function parseXHTMLString(self, s) {
+        return self._xmlparser.parseXHTMLString(s);
+    },
+
+    function appendNodeContent(self, node, innerHTML) {
+        var doc = self.parseXHTMLString(innerHTML);
+        node.appendChild(document.importNode(doc.documentElement, true));
+    },
+
     function makeHTTPRequest(self) {
         return new XMLHttpRequest();
     });
@@ -722,79 +1111,6 @@ Divmod.Runtime.InternetExplorer.methods(
         }
     },
 
-    function firstNodeByAttribute(self, root, attrName, attrValue) {
-        /* duplicate this here rather than adding an "onlyOne" arg to
-           nodesByAttribute so adding an extra arg accidentally doesn't change
-           it's behaviour if called directly
-        */
-        var descend = Divmod.Runtime.Platform.DOM_DESCEND;
-        var terminate = Divmod.Runtime.Platform.DOM_TERMINATE;
-
-        var result = null;
-        self.traverse(
-            root,
-            function(node) {
-                if (self.getAttribute(node, attrName) == attrValue) {
-                    result = node;
-                    return terminate;
-                }
-                return descend;
-            });
-        if (result === null) {
-            throw new Error("Failed to discover node with " + attrName +
-                            " value " + attrValue + " beneath " + root +
-                            " (programmer error).");
-        }
-        return result;
-    },
-
-    function nodesByAttribute(self, root, attrName, attrValue) {
-        var descend = Divmod.Runtime.Platform.DOM_DESCEND;
-        var results = [];
-        self.traverse(
-            root,
-            function(node) {
-                if (self.getAttribute(node, attrName) == attrValue) {
-                    results.push(node);
-                }
-                return descend;
-            });
-        return results;
-    },
-
-    function nodeByAttribute(self, root, attrName, attrValue, /* optional */ defaultNode) {
-        var nodes = self.nodesByAttribute(root, attrName, attrValue);
-        if (nodes.length > 1) {
-            throw new Error("Found too many " + attrName + " = " + attrValue);
-        } else if (nodes.length < 1) {
-            if (defaultNode === undefined) {
-                throw new Error("Failed to discover node with class value " +
-                                attrValue + " beneath " + root +
-                                " (programmer error).");
-            } else {
-                return defaultNode;
-            }
-
-        } else {
-            var result = nodes[0];
-            return result;
-        }
-    },
-
-    function loadScript(self, location) {
-        // The <script> tricks the base implementation uses produce
-        // spectacularly bizarre behaviour in IE, so we just use getPage/eval
-        // here.
-
-        var req = Divmod.Runtime.theRuntime.getPage(location);
-        var d = req[1];
-        d.addCallback(
-            function (result) {
-                eval(result['response']);
-            });
-        return d;
-    },
-
     /**
      * Simulate importNode by rebuilding the node contents in a node created in
      * the browser document; IE doesn't support DOM Level 2, and importNode in
@@ -826,19 +1142,9 @@ Divmod.Runtime.InternetExplorer.methods(
             var DOCUMENT_FRAGMENT_NODE = 11;
             if (root.nodeType == DOCUMENT_FRAGMENT_NODE) {
                 // We're in a DocumentFragment, and thus getElementById won't
-                // find any of our children. Let's insert ourselves into the
-                // document temporarily.
-                var currentParent = node.parentNode;
-                var nextSibling = node.nextSibling;
-                currentParent.removeChild(node);
-                document.documentElement.appendChild(node);
-
-                // Try again
-                foundNode = node.ownerDocument.getElementById(id);
-
-                // And now put us back where we used to be.
-                node.parentNode.removeChild(node);
-                currentParent.insertBefore(node, nextSibling);
+                // find any of our children.
+                foundNode = Divmod.Runtime._getElementByIdWithDocument(
+                    node, document, id);
             }
         }
 
@@ -850,7 +1156,7 @@ Divmod.Runtime.InternetExplorer.methods(
     });
 
 
-Divmod.Runtime.Opera = Divmod.Runtime.Platform.subclass("Divmod.Runtime.Opera");
+Divmod.Runtime.Opera = Divmod.Runtime.XPathSupportingPlatform.subclass("Divmod.Runtime.Opera");
 
 Divmod.Runtime.Opera.isThisTheOne = function isThisTheOne() {
     try {
@@ -912,39 +1218,14 @@ Divmod.Runtime.Opera.methods(
     });
 
 
-Divmod.Runtime.SpidermonkeyStandalone = Divmod.Class.subclass();
-Divmod.Runtime.SpidermonkeyStandalone.isThisTheOne = function() {
-    try {
-        navigator;
-    } catch (err) {
-        if (!(err instanceof ReferenceError)) {
-            throw err;
-        }
-        try {
-            document;
-        } catch (err) {
-            if (!(err instanceof ReferenceError)) {
-                throw err;
-            }
-            try {
-                window;
-            } catch (err) {
-                if (!(err instanceof ReferenceError)) {
-                    throw err;
-                }
-                return true;
-            }
-        }
-    }
-    return false;
-};
 
 Divmod.Runtime.Platform.determinePlatform = function determinePlatform() {
     var platforms = [
+        Divmod.Runtime.WebKit,
         Divmod.Runtime.Firefox,
         Divmod.Runtime.InternetExplorer,
         Divmod.Runtime.Opera,
-        Divmod.Runtime.SpidermonkeyStandalone];
+        Divmod.Runtime.Spidermonkey];
     for (var cls = 0; cls < platforms.length; ++cls) {
         if (platforms[cls].isThisTheOne()) {
             return platforms[cls];
